{
Copyright 2000 (c) Aurlien GATEAU aka Bug*Killer
aurelien.gateau@mail.dotcom.fr

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
}
Unit Vector;


interface
    uses SysUtils, Windows;

	const MaxVectorSize = Maxint div 16;
    const SVectorIndexError = 'Index de liste hors limites (%d)';
    const SVectorCapacityError = 'Dpassement de capacit';
    const SVectorCountError = 'Valeur de taille non valide';

type
    EVectorError = class(Exception);
    PObjectArray = ^TObjectArray;
    TObjectArray = array[0..MaxVectorSize - 1] of TObject;


    TVector = class(TObject)
    protected
        FItems: PObjectArray;
        FCount: Integer;
        FCapacity: Integer;

        function Get(Index: Integer): TObject;
        procedure Grow; virtual;
        procedure Put(Index: Integer; Item: TObject);
        procedure SetCapacity(NewCapacity: Integer);
        procedure SetCount(NewCount: Integer);
    public
        destructor Destroy; override;
        function Add(Item: TObject): Integer;
        procedure DeleteAll; dynamic;
        procedure Delete(Index: Integer);
        procedure Exchange(Index1, Index2: Integer);
        function First: TObject;
        function Last: TObject;
        function IndexOf(Item: TObject): Integer;
        procedure Insert(Index: Integer; Item: TObject);

        property Capacity: Integer read FCapacity write SetCapacity;
        property Count: Integer read FCount write SetCount;
        property Items[Index: Integer]: TObject read Get write Put; default;

        class procedure Error(const Msg: string; Data: Integer); virtual;
    end;

implementation


{-Methods-------------------------------------------------------------}
destructor TVector.Destroy;
begin
    DeleteAll;
end;


function TVector.Add(Item: TObject): Integer;
begin
    Result := FCount;

    if FCount = FCapacity then Grow;
    FItems^[FCount] := Item;
    Inc(FCount);
end;


procedure TVector.DeleteAll;
begin
    SetCount(0);
    SetCapacity(0);
end;

procedure TVector.Delete(Index: Integer);
begin
    if (Index < 0) or (Index >= FCount) then Error(SVectorIndexError, Index);

    FItems^[Index].Free;

    Dec(FCount);
    if Index < FCount then
        System.Move(FItems^[Index + 1], FItems^[Index],(FCount - Index) * SizeOf(TObject));
end;


procedure TVector.Exchange(Index1, Index2: Integer);
var
    Item: TObject;
begin
    if (Index1 < 0) or (Index1 >= FCount) then Error(SVectorIndexError, Index1);
    if (Index2 < 0) or (Index2 >= FCount) then Error(SVectorIndexError, Index2);
    Item := FItems^[Index1];
    FItems^[Index1] := FItems^[Index2];
    FItems^[Index2] := Item;
end;


function TVector.Get(Index: Integer): TObject;
begin
    if (Index < 0) or (Index >= FCount) then Error(SVectorIndexError, Index);
    Result := FItems^[Index];
end;

procedure TVector.Put(Index: Integer; Item: TObject);
begin
    if (Index < 0) or (Index >= FCount) then Error(SVectorIndexError, Index);
    FItems^[Index] := Item;
end;


procedure TVector.Grow;
var
    Delta: Integer;
begin
    if FCapacity > 64 then Delta := FCapacity div 4 else
        if FCapacity > 8 then Delta := 16 else
            Delta := 4;

    SetCapacity(FCapacity + Delta);
end;


function TVector.IndexOf(Item: TObject): Integer;
begin
    Result := 0;
    while (Result < FCount) and (FItems^[Result] <> Item) do Inc(Result);
    if Result = FCount then Result := -1;
end;


procedure TVector.Insert(Index: Integer; Item: TObject);
begin
    if (Index < 0) or (Index > FCount) then Error(SVectorIndexError, Index);
    if FCount = FCapacity then Grow;
    if Index < FCount then
        System.Move(FItems^[Index], FItems^[Index + 1],(FCount - Index) * SizeOf(TObject));

    FItems^[Index] := Item;
    Inc(FCount);
end;


function TVector.First: TObject;
begin
    Result := Get(0);
end;

function TVector.Last: TObject;
begin
    Result := Get(FCount - 1);
end;


procedure TVector.SetCapacity(NewCapacity: Integer);
begin
// Check param
    if (NewCapacity < FCount) or (NewCapacity > MaxVectorSize) then Error(SVectorCapacityError, NewCapacity);

    if NewCapacity <> FCapacity then
    begin
        ReallocMem(FItems, NewCapacity * SizeOf(TObject));
        FCapacity := NewCapacity;
    end;
end;


procedure TVector.SetCount(NewCount: Integer);
var
    Num:LongInt;
begin
// Check param
    if (NewCount < 0) or (NewCount > MaxVectorSize) then
        Error(SVectorCountError, NewCount);

// Increase capacity if necessary
    if NewCount > FCapacity then SetCapacity(NewCount);

// Create or delete items
    if NewCount > FCount then
    // Fill new items with nil
    begin
        for Num:=NewCount to FCount-1 do FItems^[Num]:=nil;
    end
    else
    begin
        if NewCount < FCount then
        // Delete extra items
        begin
            for Num:=NewCount to FCount-1 do FItems^[Num].Free;
        end;
    end;

    FCount := NewCount;
end;


{-Class methods-------------------------------------------------------}
class procedure TVector.Error(const Msg: string; Data: Integer);

    function ReturnAddr: Pointer;
    asm
        MOV     EAX,[EBP+4]
    end;

begin
    raise EVectorError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;
end.
